/* TinyDecisionTreeClassifier library
 * Copyright (c) 2023-2024 Aleksei Karavaev
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef DECISION_TREE_CLASSIFIER_CPP
#define DECISION_TREE_CLASSIFIER_CPP
#include "TinyDecisionTreeClassifier.h"

template < typename T >
TinyDecisionTreeClassifier<T>::TinyDecisionTreeClassifier(uint16_t maxDepth, uint16_t minSamplesSplit, Goodness_criterion gc){
    this->maxDepth=maxDepth;
    this->minSamplesSplit=minSamplesSplit;
    this->root = NULL;
    this->gc = gc;
};

template < typename T >
TinyDecisionTreeClassifier<T>::~TinyDecisionTreeClassifier(){
    if(root!=NULL){
        root->cleanup();
        delete root;   
    }
};

// number of rows is the same for both
template < typename T >
void TinyDecisionTreeClassifier<T>::fit(T** X,T** Y, uint32_t rows,uint32_t cols,RowsSubIndexes* rootRsi){
    if(root!=NULL){
        root->cleanup();
        delete root;   
        root=NULL;
    }
    root = new Node();
    bool cleanRsi=false;
    if(rootRsi==NULL){
        cleanRsi=true;
        rootRsi = new RowsSubIndexes;
        rootRsi->size = rows;
        rootRsi->indexes = (uint32_t *)malloc(rows*sizeof(uint32_t));
        for(uint32_t i=0;i<rows;i++){
            rootRsi->indexes[i]=i;
        }
    }
    countUniqueValuesAndOccurances(Y,rootRsi,0,&uniqueClassValuesAndOccurances);
    recurcisiveFit(root,X,Y,rootRsi,cols,1);
    this->cols = cols;
    if(cleanRsi)delete rootRsi;
};

template < typename T >
void TinyDecisionTreeClassifier<T>::plot(void){
    plot(root,0);
}

template < typename T >
void TinyDecisionTreeClassifier<T>::plot(Node* node,uint32_t depth){
    if(node->thReady){
        for(uint32_t i=0;i<depth;i++){
            DTR_PLOT_PRINT("               ");
        }
        DTR_PLOT_PRINT("if(X[");
        DTR_PLOT_PRINT_U(node->nodeThColumn);
        DTR_PLOT_PRINT("]<=");
        DTR_PLOT_PRINT_F(node->nodeTh);
        DTR_PLOT_PRINTLN(")");
    }
    if(node->decisionReady){
        for(uint32_t i=0;i<depth;i++){
            DTR_PLOT_PRINT("               ");
        }
        DTR_PLOT_PRINT("Y=(");
        DTR_PLOT_PRINT_F(node->decision);
        DTR_PLOT_PRINTLN(")");
    }
    
    if(node->children[0]!=NULL){
        for(uint16_t i=0;i<2;i++){
            plot(node->children[i],depth+1);
        }
    }
}

template < typename T >
T TinyDecisionTreeClassifier<T>::predict(T* X){
    if(root==NULL)return 0;
    return root->decide(X);
}

template < typename T >
float TinyDecisionTreeClassifier<T>::score(T** X,T** Y,uint32_t rows){
    float score = 0;
    for(uint32_t i=0;i<rows;i++){
        if(predict(X[i])==Y[i][0])score++;
    }
    return score/rows;
}

template < typename T >
TinyDecisionTreeClassifier<T>::Node::Node(){}

template < typename T >
int16_t TinyDecisionTreeClassifier<T>::recurcisiveFit(Node* node,T** X,T** Y, RowsSubIndexes* rsi, uint32_t cols, uint32_t currentDepth){
    DTR_DEBUG_PRINTLN();
    DTR_DEBUG_PRINT("Recursive fit for node:");
    DTR_DEBUG_PRINTLN((uint32_t)this);
    DTR_DEBUG_PRINTLN("Input data:");
    for(uint32_t i=0;i<rsi->size;i++){
        for(uint32_t j=0;j<cols;j++){
            DTR_DEBUG_PRINT(X[rsi->indexes[i]][j]);
            DTR_DEBUG_PRINT(' ');
        }
        DTR_DEBUG_PRINT("-> ");
        DTR_DEBUG_PRINTLN(Y[rsi->indexes[i]][0]);
    }
    countOccurances(Y,rsi,0,&uniqueClassValuesAndOccurances);
    //check only class
    if(onlyClassLeft(&uniqueClassValuesAndOccurances)==1){
        node->decision = getMajorClass(&uniqueClassValuesAndOccurances);
        node->decisionReady = true;
        DTR_DEBUG_PRINT("Only unique class, finish splitting, decision is ");
        DTR_DEBUG_PRINTLN(node->decision);
        return 0;
    }
    else if (rsi->size<minSamplesSplit){
        node->decision = getMajorClass(&uniqueClassValuesAndOccurances);
        node->decisionReady = true;
        DTR_DEBUG_PRINTLN("Length of X and Y is less than min sample split, decision is ");
        DTR_DEBUG_PRINTLN(node->decision);
        return 0;
    }
    else if (currentDepth == maxDepth){
        node->decision = getMajorClass(&uniqueClassValuesAndOccurances);
        node->decisionReady = true;
        DTR_DEBUG_PRINT("Max depth reached, decision is ");
        DTR_DEBUG_PRINTLN(node->decision);
        return 0;
    }
 
    RowsSubIndexes rsiAboveTh;
    RowsSubIndexes rsiBelowTh;
    rsiAboveTh.indexes=(uint32_t*)malloc((rsi->size)*sizeof(uint32_t));
    rsiBelowTh.indexes=(uint32_t*)malloc((rsi->size)*sizeof(uint32_t));

    T th;
    uint32_t thColumn;
    int16_t rslt;
    switch (gc)
    {
    case INFO_GAIN:
        rslt = getBestSplitInfoGain(X,Y,rsi,cols,&rsiAboveTh,&rsiBelowTh,&th,&thColumn);
        break;
    case GINI:
        rslt = getBestSplitGini(X,Y,rsi,cols,&rsiAboveTh,&rsiBelowTh,&th,&thColumn);
        break;
    default:
        rslt = getBestSplitInfoGain(X,Y,rsi,cols,&rsiAboveTh,&rsiBelowTh,&th,&thColumn);
        break;
    }


    if(rslt != CANT_SPLIT_ALL_THE_SAMPLES_HAVE_THE_SAME_VALUE){
        node->children[0] = new Node();
        node->children[1] = new Node();

        node->nodeTh=th;
        node->nodeThColumn=thColumn;
        node->thReady = true;
        node->decision = getMajorClass(&uniqueClassValuesAndOccurances);

        recurcisiveFit(node->children[0],X,Y,&rsiBelowTh,cols,currentDepth+1);
        recurcisiveFit(node->children[1],X,Y,&rsiAboveTh,cols,currentDepth+1);
    }else{
        node->decision = getMajorClass(&uniqueClassValuesAndOccurances);
        node->decisionReady = true;
    }
    return 0;
}

template < typename T >
float TinyDecisionTreeClassifier<T>::computeEntropy(T** Y, RowsSubIndexes* rsi){
    float entropy=0;
    if(rsi->size<2){
        return 0;
    }
    else{
        countOccurances(Y,rsi,0,&uniqueClassValuesAndOccurances);
        uint32_t i;
        for(i = 0;i<uniqueClassValuesAndOccurances.uniqueValuesSize;i++){
            float freq = ((float)(uniqueClassValuesAndOccurances.uniqueValuesOccurances[i]))/rsi->size;
            entropy += -(freq*(log(freq+1e-6)/log(2)));
        }
    }
    return entropy;
}

template < typename T >
int16_t TinyDecisionTreeClassifier<T>::getBestSplitInfoGain(T** X,T** Y, RowsSubIndexes* rsi, uint32_t cols, RowsSubIndexes* rsiAboveTh, RowsSubIndexes* rsiBelowTh, T* threshold, uint32_t* column){
    float entropyBeforeTheSplit=computeEntropy(Y,rsi);
    float entropyAbove;
    float entropyBelow;
    float infoGain;//A few operations can be saved by using Entropy directly without computing Y entropy
    uint32_t k=0;

    //finding max
    float bestInfoGain=-FLT_MAX;
    RowsSubIndexes currentRsiBelowTh;
    RowsSubIndexes currentRsiAboveTh;
    currentRsiBelowTh.indexes=(uint32_t *)malloc(rsi->size*sizeof(uint32_t));
    currentRsiAboveTh.indexes=(uint32_t *)malloc(rsi->size*sizeof(uint32_t));

    for(uint32_t i=0;i<cols;i++){
        //Sort the column
        T* sorted = (T *)malloc(rsi->size*sizeof(T));
        uint32_t* idxs = (uint32_t *)malloc(rsi->size*sizeof(uint32_t));

        for(uint32_t j=0;j<rsi->size;j++){
            sorted[j]=X[rsi->indexes[j]][i];
            idxs[j]=rsi->indexes[j];
        }
        qsort(sorted,idxs,rsi->size);

        //Find where the values differ and check entropy (to find max for threshold generation)
        for(uint32_t j=1;j<rsi->size;j++){
            if(sorted[j]!=sorted[j-1]){
                currentRsiBelowTh.size=j;
                for(uint32_t k=0;k<j;k++){
                    currentRsiBelowTh.indexes[k]=idxs[k];
                }

                entropyBelow = computeEntropy(Y,&currentRsiBelowTh);
                currentRsiAboveTh.size=(rsi->size-j);
                for(uint32_t k=j;k<(rsi->size);k++){
                    currentRsiAboveTh.indexes[k-j]=idxs[k];
                }
                entropyAbove = computeEntropy(Y,&currentRsiAboveTh);

                infoGain = entropyBeforeTheSplit - (entropyBelow*(((float)j)/((float)rsi->size))+entropyAbove*(((float)(rsi->size-j))/((float)rsi->size)));
                if(infoGain>bestInfoGain){
                    bestInfoGain=infoGain;
                    *threshold=(sorted[j-1] + sorted[j])/2;
                    *column=i;

                    // if(rsiBelowTh->indexes)free(rsiBelowTh->indexes);
                    // rsiBelowTh->indexes=(uint32_t*)malloc(j*sizeof(uint32_t));
                    rsiBelowTh->size=j;
                    for(k=0;k<j;k++){
                        rsiBelowTh->indexes[k]=idxs[k];
                    }

                    // if(rsiAboveTh->indexes)free(rsiAboveTh->indexes);
                    // rsiAboveTh->indexes=(uint32_t*)malloc((rsi->size-j)*sizeof(uint32_t));
                    rsiAboveTh->size=(rsi->size-j);
                    for(k=j;k<(rsi->size);k++){
                        rsiAboveTh->indexes[k-j]=idxs[k];
                    }
                }
            }
        }
        free(sorted);
        free(idxs);
    }
    if(bestInfoGain==-FLT_MAX){
        DTR_DEBUG_PRINT("Can't split all the samples have the same value");
        return CANT_SPLIT_ALL_THE_SAMPLES_HAVE_THE_SAME_VALUE;
    }
    DTR_DEBUG_PRINT("Best split in column ");
    DTR_DEBUG_PRINT(*column);
    DTR_DEBUG_PRINT(" with threhold ");
    DTR_DEBUG_PRINT(*threshold);
    DTR_DEBUG_PRINT(" and infogain ");
    DTR_DEBUG_PRINTLN(bestInfoGain);
    DTR_DEBUG_PRINT("Data below thehold ");
    for(uint32_t k=0;k<rsiBelowTh->size;k++){
        DTR_DEBUG_PRINT(" ");
        DTR_DEBUG_PRINT(X[rsiBelowTh->indexes[k]][*column]);
    }
    DTR_DEBUG_PRINTLN();
    DTR_DEBUG_PRINT("Data above thehold ");
    for(uint32_t k=0;k<rsiAboveTh->size;k++){
        DTR_DEBUG_PRINT(" ");
        DTR_DEBUG_PRINT(X[rsiAboveTh->indexes[k]][*column]);
    }
    DTR_DEBUG_PRINTLN();
    return 0;
}

template < typename T >
float TinyDecisionTreeClassifier<T>::computeGini(T** Y,RowsSubIndexes* rsi){
    float gini=1;
    if(rsi->size<2){
        return 0;
    }
    else{
        countOccurances(Y,rsi,0,&uniqueClassValuesAndOccurances);
        uint32_t i;
        for(i = 0;i<uniqueClassValuesAndOccurances.uniqueValuesSize;i++){
            float freq = ((float)(uniqueClassValuesAndOccurances.uniqueValuesOccurances[i]))/rsi->size;
            gini -= (float)((float)freq*(float)freq);
        }
    }
    return gini;
}

template < typename T >
int16_t TinyDecisionTreeClassifier<T>::getBestSplitGini(T** X,T** Y, RowsSubIndexes* rsi, uint32_t cols, RowsSubIndexes* rsiAboveTh, RowsSubIndexes* rsiBelowTh, T* threshold, uint32_t* column){
    float giniAbove;
    float giniBelow;
    float giniSplit;//A few operations can be saved by using Entropy directly without computing Y entropy
    uint32_t k;

    //finding max
    float bestGiniSplit=0.5;
    struct RowsSubIndexes currentRsiBelowTh;
    struct RowsSubIndexes currentRsiAboveTh;
    currentRsiBelowTh.indexes=(uint32_t *)malloc(rsi->size*sizeof(uint32_t));
    currentRsiAboveTh.indexes=(uint32_t *)malloc(rsi->size*sizeof(uint32_t));

    // uint32_t totalSorting=0;
    // uint32_t total=micros();
    // uint32_t totalGini=0;

    uint32_t tic=0;
    for(uint32_t i=0;i<cols;i++){
        T* sorted = (T *)malloc(rsi->size*sizeof(T));
        uint32_t* idxs = (uint32_t *)malloc(rsi->size*sizeof(uint32_t));

        for(uint32_t j=0;j<rsi->size;j++){
            sorted[j]=X[rsi->indexes[j]][i];
            idxs[j]=rsi->indexes[j];
        }
        qsort(sorted,idxs,rsi->size);

        //Find where the values differ and check entropy (to find max for threshold generation)
        for(uint32_t j=1;j<rsi->size;j++){
            if(sorted[j]!=sorted[j-1]){
                currentRsiBelowTh.size=j;
                for(k=0;k<j;k++){
                    currentRsiBelowTh.indexes[k]=idxs[k];
                }
                giniBelow = computeGini(Y,&currentRsiBelowTh);
                currentRsiAboveTh.size=(rsi->size-j);
                for(k=j;k<(rsi->size);k++){
                    currentRsiAboveTh.indexes[k-j]=idxs[k];
                }

                giniAbove = computeGini(Y,&currentRsiAboveTh);

                giniSplit = (giniBelow*(((float)j)/((float)rsi->size))+giniAbove*(((float)(rsi->size-j))/((float)rsi->size)));
                if(giniSplit<bestGiniSplit){
                    bestGiniSplit=giniSplit;
                    *threshold=(sorted[j-1] + sorted[j])/2;
                    *column=i;

                    rsiBelowTh->size=j;

                    for(k=0;k<j;k++){
                        rsiBelowTh->indexes[k]=idxs[k];
                    }

                    rsiAboveTh->size=(rsi->size-j);
                    for(k=j;k<(rsi->size);k++){
                        rsiAboveTh->indexes[k-j]=idxs[k];
                    }
                }
            }
        }
        free(sorted);
        free(idxs);
    }

    if(bestGiniSplit==0.5){
        DTR_DEBUG_PRINT("Can't split all the samples have the same value");
        return CANT_SPLIT_ALL_THE_SAMPLES_HAVE_THE_SAME_VALUE;
    }
    DTR_DEBUG_PRINT("Best split in column ");
    DTR_DEBUG_PRINT(*column);
    DTR_DEBUG_PRINT(" with threhold ");
    DTR_DEBUG_PRINT(*threshold);
    DTR_DEBUG_PRINT(" and gini ");
    DTR_DEBUG_PRINT(bestGiniSplit);
    DTR_DEBUG_PRINTLN("");
    DTR_DEBUG_PRINT("Data below thehold ");
    for(k=0;k<rsiBelowTh->size;k++){
        DTR_DEBUG_PRINT(" ");
        DTR_DEBUG_PRINT(X[rsiBelowTh->indexes[k]][*column]);
    }
    DTR_DEBUG_PRINTLN("");
    DTR_DEBUG_PRINT("Data above thehold ");
    for(k=0;k<rsiAboveTh->size;k++){
        DTR_DEBUG_PRINTLN(" ");
        DTR_DEBUG_PRINT(X[rsiAboveTh->indexes[k]][*column]);
    }
    DTR_DEBUG_PRINTLN("");
    return 0;
}

template < typename T >
void TinyDecisionTreeClassifier<T>::countUniqueValuesAndOccurances(T ** ar, RowsSubIndexes* rsi, uint32_t column, UniqueValues* uv){
    T* sorted = (T *)malloc(rsi->size*sizeof(T));

    for(uint32_t i=0;i<rsi->size;i++){
        sorted[i]=ar[rsi->indexes[i]][column];
    }
    qsort(sorted,rsi->size);
    
    uv->uniqueValues = (T *)malloc(1*sizeof(T));
    uv->uniqueValuesOccurances = (uint32_t *)malloc(1*sizeof(uint32_t));
    uv->uniqueValues[0] = sorted[0];
    uv->uniqueValuesSize = 1;
    uv->uniqueValuesOccurances[0]=1;

    for (uint16_t i=1;i<rsi->size;i++){
        if(sorted[i]!=sorted[i-1]){
            uv->uniqueValues = (T *)realloc(uv->uniqueValues,(uv->uniqueValuesSize+1)*sizeof(T));
            uv->uniqueValuesOccurances = (uint32_t *)realloc(uv->uniqueValuesOccurances,(uv->uniqueValuesSize+1)*sizeof(uint32_t));
            uv->uniqueValuesOccurances[uv->uniqueValuesSize]=1;
            uv->uniqueValues[uv->uniqueValuesSize]=sorted[i];
            uv->uniqueValuesSize++;
        }else{
            uv->uniqueValuesOccurances[uv->uniqueValuesSize-1]++;
        }
    }
    free(sorted);
}

template < typename T >
void TinyDecisionTreeClassifier<T>::countOccurances(T ** ar, RowsSubIndexes* rsi, uint32_t column, UniqueValues* uv){
    uint16_t i=0;
    uint16_t j=0;
    for(j=0;j<uv->uniqueValuesSize;j++){
        uv->uniqueValuesOccurances[j]=0;
    }
    for (i=0;i<rsi->size;i++){
        for(j=0;j<uv->uniqueValuesSize;j++){
            if(ar[rsi->indexes[i]][column]==uv->uniqueValues[j])uv->uniqueValuesOccurances[j]++;
        }
    }
}

template < typename T >
bool TinyDecisionTreeClassifier<T>::onlyClassLeft(UniqueValues* uv){
    uint32_t i;
    uint32_t non0Ctr=0;
    for(i=0;i<uv->uniqueValuesSize;i++){
        if(uv->uniqueValuesOccurances[i]!=0)non0Ctr++; 
    }
    return non0Ctr==1;
}

template < typename T >
void TinyDecisionTreeClassifier<T>::qsort(T *ar, uint32_t n) 
{
    if (n < 2)
        return;
    T p = ar[n / 2];
    T *l = ar;
    T *r = ar + n - 1;
    while (l <= r) {
    if (*l < p) {
        l++;
    }
    else if (*r > p) {
        r--;
    }
    else {
        T t = *l;
        *l = *r;
        *r = t;
        l++;
        r--;
    }
    }
    qsort(ar, r - ar + 1);
    qsort(l, ar + n - l);
}

template < typename T >
void TinyDecisionTreeClassifier<T>::qsort(T *ar, uint32_t *idx, uint32_t n) 
{
    if (n < 2)
        return;
    T p = ar[n / 2];
    T *l = ar;
    T *r = ar + n - 1;
    uint32_t *li = idx;
    uint32_t *ri = idx + n - 1;


    while (l <= r) {
    if (*l < p) {
        l++;
        li++;
    }
    else if (*r > p) {
        r--;
        ri--;
    }
    else {
        T t = *l;
        *l = *r;
        *r = t;
        l++;
        r--;
        uint32_t ti = *li;
        *li = *ri;
        *ri = ti;
        li++;
        ri--;
    }
    }
    qsort(ar, idx, r - ar + 1);
    qsort(l, li, ar + n - l);
}

template < typename T >
T TinyDecisionTreeClassifier<T>::getMajorClass(UniqueValues* uv){
    uint32_t maxOcc=0;
    uint32_t maxOccIdx=0; 
    uint32_t i;
    for(i=0;i<uv->uniqueValuesSize;i++){
        if(uv->uniqueValuesOccurances[i]>maxOcc){
            maxOcc=uv->uniqueValuesOccurances[i];
            maxOccIdx=i;
        }
    }
    return uv->uniqueValues[maxOccIdx];
}

template < typename T >
T TinyDecisionTreeClassifier<T>::Node::decide(T* X){
    if(decisionReady)return decision;
    else{
        if(thReady){
            if(children[0]!=NULL){
                if(X[nodeThColumn]<=nodeTh){
                    return children[0]->decide(X);
                }else{
                    return children[1]->decide(X);
                }
            }
            else{
                DTR_DEBUG_PRINTLN("Warning node have no children and no decision,returning 0");
                return 0;
            }
        }
        else{
            DTR_DEBUG_PRINTLN("Warning node have no threhold and no decision,returning 0");
            return 0;
        }
    }
}

template < typename T >
void TinyDecisionTreeClassifier<T>::Node::decideBasedOnPreLeafNode(T *X,NodeDecisionPair* ndp){
    if(decisionReady){
        ndp->decision = decision;
        ndp->nodePtr = NULL;
    }
    else{
        if(thReady){
            if(children[0]!=NULL && children[1]!=NULL){
                if(children[0]->children[0] == NULL && children[0]->children[1] == NULL && children[1]->children[0] == NULL && children[1]->children[1] == NULL){
                    DTR_DEBUG_PRINT("Node ");
                    DTR_DEBUG_PRINT((uint32_t) this);
                    DTR_DEBUG_PRINT("(Y=");
                    DTR_DEBUG_PRINT(decision);
                    DTR_DEBUG_PRINT(",th=");
                    DTR_DEBUG_PRINT(nodeTh);
                    DTR_DEBUG_PRINT("): ");
                    // #endif
                    ndp->decision = decision;
                    ndp->nodePtr = this;
                }else{
                    if(children[0]!=NULL){
                        if(X[nodeThColumn]<=nodeTh){
                            children[0]->decideBasedOnPreLeafNode(X,ndp);
                        }else{
                            children[1]->decideBasedOnPreLeafNode(X,ndp);
                        }                
                    }
                }
            }else{
                DTR_DEBUG_PRINTLN("Warning, node has no children and no decision, or just one child");
            }
        }
        else{
            DTR_DEBUG_PRINTLN("Warning, node has no threhold and no decision");
        }
    }
}

template < typename T >
void TinyDecisionTreeClassifier<T>::Node::cleanup(){
    if(children[0]!=NULL){
        for(uint32_t i=0;i<2;i++){
            children[i]->cleanup();
            delete children[i];
            decisionReady=true;
            thReady=false;
            children[i] = NULL;
        }
    }
}

template < typename T >
T TinyDecisionTreeClassifier<T>::predictBasedOnPreLeafNode(T *X,typename Node::NodeDecisionPair* ndp){
    if(root==NULL)return 0;
    root->decideBasedOnPreLeafNode(X,ndp);
    return 0;
}

template < typename T >
bool TinyDecisionTreeClassifier<T>::prune(T** X,T** Y,uint32_t rows){
    bool prunedSuccess = false;
    Node** nodesUnderPruning = NULL;
    int32_t* nodesUnderPruningLeavesScore = NULL;
    uint32_t nodesUnderPruningLen = 0;
    DTR_DEBUG_PRINTLN("Cost-complexity pruning start");

    for(uint32_t i=0;i<rows;i++){
        typename Node::NodeDecisionPair ndp;
        DTR_DEBUG_PRINT("Pruner is trying to classify: X=[");
        uint32_t k;
        for(k=0;k<cols;k++){
            DTR_DEBUG_PRINT(X[i][k]);
            DTR_DEBUG_PRINT(",");
        }
        DTR_DEBUG_PRINT("], Y=");
        DTR_DEBUG_PRINT(Y[i][0]);
        DTR_DEBUG_PRINT(", ");
        predictBasedOnPreLeafNode(X[i],&ndp);
        if(ndp.nodePtr!=NULL){
            if(nodesUnderPruning == NULL){
                nodesUnderPruningLen++;
                nodesUnderPruning = (Node **)malloc(nodesUnderPruningLen*sizeof(Node*));
                nodesUnderPruningLeavesScore = (int32_t *)malloc(nodesUnderPruningLen*sizeof(int32_t));

                nodesUnderPruning[0] = ndp.nodePtr;
                nodesUnderPruningLeavesScore[0] = 0;
                
                if(ndp.decision == Y[i][0]){
                    DTR_DEBUG_PRINT("OK,");
                    nodesUnderPruningLeavesScore[0]--;
                    }
                else{
                    DTR_DEBUG_PRINT("ER,");
                }
                DTR_DEBUG_PRINT(" Unpruned: ");

                if(predict(X[i])==Y[i][0]){
                    DTR_DEBUG_PRINTLN("OK");
                    nodesUnderPruningLeavesScore[0]++;
                }
                else{
                    DTR_DEBUG_PRINTLN("ER");                    
                }
            }
            else{
                bool nodeFound = false;
                uint32_t j;
                for(j=0;j<nodesUnderPruningLen;j++){
                    if(nodesUnderPruning[j] == ndp.nodePtr){
                        if(ndp.decision == Y[i][0]){
                            DTR_DEBUG_PRINT("OK,");
                            nodesUnderPruningLeavesScore[j]--;
                            }
                        else{
                            DTR_DEBUG_PRINT("ER,");
                        }
                        DTR_DEBUG_PRINT(" Unpruned: ");

                        if(predict(X[i])==Y[i][0]){
                            DTR_DEBUG_PRINTLN("OK");
                            nodesUnderPruningLeavesScore[j]++;
                        }
                        else{
                            DTR_DEBUG_PRINTLN("ER");                    
                        }
                        nodeFound = true;
                    }
                }
                if(!nodeFound){
                    nodesUnderPruningLen++;
                    nodesUnderPruning = (Node **)realloc(nodesUnderPruning,nodesUnderPruningLen*sizeof(Node*));
                    nodesUnderPruningLeavesScore = (int32_t *)realloc(nodesUnderPruningLeavesScore,nodesUnderPruningLen*sizeof(int32_t));

                    nodesUnderPruning[nodesUnderPruningLen-1] = ndp.nodePtr;
                    nodesUnderPruningLeavesScore[nodesUnderPruningLen-1] = 0;

                    if(ndp.decision == Y[i][0]){
                        DTR_DEBUG_PRINT("OK,");
                        nodesUnderPruningLeavesScore[nodesUnderPruningLen-1]--;
                        }
                    else{
                        DTR_DEBUG_PRINT("ER,");
                    }
                    DTR_DEBUG_PRINT(" Unpruned: ");
    
                    if(predict(X[i])==Y[i][0]){
                        DTR_DEBUG_PRINTLN("OK");
                        nodesUnderPruningLeavesScore[nodesUnderPruningLen-1]++;
                    }
                    else{
                        DTR_DEBUG_PRINTLN("ER");                    
                    }
                }
            }
        }else{
        DTR_DEBUG_PRINTLN(" OK, Unpruned  OK");
        }
    }
    uint32_t i=0;
    DTR_DEBUG_PRINTLN("Tree evaluation for pruning finished");
    for(i=0;i<nodesUnderPruningLen;i++){
        DTR_DEBUG_PRINT("Node:")
        DTR_DEBUG_PRINT((uint32_t)nodesUnderPruning[i]);
        DTR_DEBUG_PRINT(" Score: ");
        DTR_DEBUG_PRINTLN(nodesUnderPruningLeavesScore[i]);
        if(nodesUnderPruningLeavesScore[i]<=0){
            prunedSuccess=true;
            DTR_DEBUG_PRINT("Delete ");
            DTR_DEBUG_PRINT((uint32_t)nodesUnderPruning[i]);
            DTR_DEBUG_PRINTLN(" node ");
            nodesUnderPruning[i]->cleanup();
        }
    }
    free(nodesUnderPruning);
    free(nodesUnderPruningLeavesScore);
    return prunedSuccess;
}

template < typename T >
int32_t TinyDecisionTreeClassifier<T>::getRootLine(Node* node,int32_t linenow){
  if(node->decisionReady){
    linenow = linenow + 1;
    return linenow;
  }
  if(node->thReady){
    int32_t lineofnode0 = getRootLine(node->children[0],linenow);
    int32_t lineofnode1 = getRootLine(node->children[1],lineofnode0);
    return (lineofnode1+1);
  }
  return linenow;
}

template < typename T >
int32_t TinyDecisionTreeClassifier<T>::nodesToChar(char* outdata, uint32_t* outdataOffset, Node* node,int32_t linenow){
  if(node->decisionReady){
    (*outdataOffset)+=sprintf(outdata+(*outdataOffset),"n,-1,%ld,-1,-1\n",(int32_t)node->decision);
    linenow = linenow + 1;
    return linenow;
  }
  if(node->thReady){
    int32_t lineofnode0 = nodesToChar(outdata,outdataOffset,node->children[0],linenow);
    int32_t lineofnode1 = nodesToChar(outdata,outdataOffset,node->children[1],lineofnode0);
    (*outdataOffset)+=sprintf(outdata+(*outdataOffset),"n,%lu,%f,%ld,%ld\n",node->nodeThColumn,(double)node->nodeTh,lineofnode0-1,lineofnode1-1);
    return (lineofnode1+1);
  }
  return linenow;
}

template < typename T >
int TinyDecisionTreeClassifier<T>::save(char** outdata, uint32_t* outdataMaxSize){
    uint32_t outdataOffset=0;
    uint32_t linenow = 0;
    int32_t rootline=0;
    if(root!=NULL){
        rootline = getRootLine(root,rootline)-1;
        outdataOffset+=sprintf(*outdata,"r,%ld\n",rootline);
        if((outdataOffset+100)>(*outdataMaxSize)){
          (*outdataMaxSize)+=100;
          *outdata = (char*)realloc(*outdata,(*outdataMaxSize));
        }
        nodesToChar(*outdata,&outdataOffset,root,linenow);
        (*outdata)[outdataOffset]=0;
    }
    return 0;
}

template < typename T >
void TinyDecisionTreeClassifier<T>::buildDescendents(Node* ptr,int32_t index,float** nodesData){
  if(nodesData[index][0]!=-1){
    ptr->decisionReady=false;
    ptr->nodeThColumn=(uint32_t)nodesData[index][0];
    ptr->nodeTh=nodesData[index][1];
    ptr->thReady=true;

    if(nodesData[index][2]!=-1){
      ptr->children[0]=(Node*)malloc(sizeof(Node));
      buildDescendents(ptr->children[0],(int32_t )nodesData[index][2],nodesData);
      }
    else ptr->children[0]=NULL;
    if(nodesData[index][3]!=-1){
      ptr->children[1]=(Node*)malloc(sizeof(Node));
      buildDescendents(ptr->children[1],(int32_t )nodesData[index][3],nodesData);
      }
    else ptr->children[1]=NULL;
  }else{
    ptr->decisionReady=true;
    ptr->decision=nodesData[index][1];
    ptr->thReady=false;
    ptr->children[0]=NULL;
    ptr->children[1]=NULL;
  }
}

template < typename T >
void TinyDecisionTreeClassifier<T>::countNodesAndRoots(uint32_t* roots,uint32_t* nodes,char* data){
  char* newline_position;
  char* line_position = data;
  int32_t rootline=0;
  int32_t feature=0;
  *nodes=0;
  *roots=0;
  while(1){
    newline_position = strchr(line_position, 'r');
    if(newline_position==NULL)break;
    sscanf(newline_position,"r,%ld",&rootline);
    (*roots)++;
    line_position = newline_position+1;
  }
  line_position = data;
  while(1){
    newline_position = strchr(line_position, 'n');
    if(newline_position==NULL)break;
    sscanf(newline_position,"n,%ld",&feature);
    (*nodes)++;
    line_position = newline_position+1;
  }
}
template < typename T >
void TinyDecisionTreeClassifier<T>::fillInNodesData(uint32_t* roots,uint32_t* nodes,char* inputdata,float** nodesData,float** rootsData){
  char* newline_position;
  char* line_position = inputdata;
  uint32_t ctr=0;
  uint32_t rootsctr=0;

  while(1){
    int32_t rootline=0;
    newline_position = strchr(line_position, 'r');
    if(newline_position==NULL)break;
    sscanf(newline_position,"r,%ld",&rootline);
    rootsData[rootsctr][0]=rootline;
    rootsctr++;
    line_position = newline_position+1;
  }
  line_position = inputdata;
  while(1){
    newline_position = strchr(line_position, 'n');
    if(newline_position==NULL)break;

    int32_t feature=0;
    float threhold=0;
    int32_t left=0;
    int32_t right=0;
    sscanf(newline_position,"n,%ld,%f,%ld,%ld",&feature,&threhold,&left,&right);
    nodesData[ctr][0]=feature;
    nodesData[ctr][1]=threhold;
    nodesData[ctr][2]=left;
    nodesData[ctr][3]=right;
    ctr++;
    line_position = newline_position+1;
  }
}

template < typename T >
int TinyDecisionTreeClassifier<T>::load(char* inputdata){
    if(root==NULL){
        root = new Node();
        uint32_t nodesSize=0;
        uint32_t rootsSize=0;
        uint32_t i=0;
        countNodesAndRoots(&rootsSize,&nodesSize,inputdata);
        if(rootsSize!=1)return -1;
        float ** nodesData = (float**)malloc(sizeof(float*)*(nodesSize));
        float ** rootsData = (float**)malloc(sizeof(float*)*(rootsSize));
        for(i=0;i<nodesSize;i++)nodesData[i] = (float*)malloc(sizeof(float)*4);
        for(i=0;i<rootsSize;i++)rootsData[i] = (float*)malloc(sizeof(float)*1);
        
        fillInNodesData(&rootsSize,&nodesSize,inputdata,nodesData,rootsData);
        buildDescendents(root,rootsData[0][0],nodesData);

        for(i=0;i<nodesSize;i++)free(nodesData[i]);
        for(i=0;i<rootsSize;i++)free(rootsData[i]);
        free(nodesData);
        free(rootsData);
        return 0;
    }
    else{
        return -1;
    }
}

#endif