
/* TinyDecisionTreeClassifier library
 * Copyright (c) 2023-2024 Aleksei Karavaev
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef DECISION_TREE_CLASSIFIER_H
#define DECISION_TREE_CLASSIFIER_H
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include "math.h"
#include "float.h"

// #define DTR_DEBUG_
#ifdef ARDUINO
    #include "Arduino.h"
    #define DTR_PLOT_PRINT(X) Serial.print(X);
    #define DTR_PLOT_PRINTLN(X) Serial.println(X);
    #define DTR_PLOT_PRINTLN_F(X) Serial.println(X,5);
    #define DTR_PLOT_PRINT_F(X)   Serial.print(X,5);
    #define DTR_PLOT_PRINT_U(X)   Serial.print(X);
    #ifdef DTR_DEBUG_
        #define DTR_DEBUG_PRINT(X) Serial.print(X);
        #define DTR_DEBUG_PRINTLN(X) Serial.println(X);
    #else
        #define DTR_DEBUG_PRINT(X);
        #define DTR_DEBUG_PRINTLN(X);
    #endif
#else
    #ifdef MBED_H
        #include "mbed.h"
        #define DTR_PLOT_PRINT(X) printf(X);
        #define DTR_PLOT_PRINTLN(X) printf(X);printf("\n\r");
        #define DTR_PLOT_PRINTLN_F(X) printf("%.5f\n\r",X);
        #define DTR_PLOT_PRINT_F(X)   printf("%.5f",X);
        #define DTR_PLOT_PRINT_U(X)   printf("%lu",X);

        #ifdef DTR_DEBUG_
            #define DTR_DEBUG_PRINT(X);
            #define DTR_DEBUG_PRINTLN(X);
        #else
            #define DTR_DEBUG_PRINT(X);
            #define DTR_DEBUG_PRINTLN(X);
        #endif

    #else
        #define DTR_DEBUG_PRINT(X);
        #define DTR_DEBUG_PRINTLN(X);
        #define DTR_PLOT_PRINT(X);
        #define DTR_PLOT_PRINTLN(X);
    #endif
#endif

#define CANT_SPLIT_ALL_THE_SAMPLES_HAVE_THE_SAME_VALUE  -1

/** @brief The main classifier class, the tempate T allows changing the datatype of the input data. Also 8-bit MCUs work faster with int8_t data type**/
template <typename T>
class TinyDecisionTreeClassifier{
    public:
        enum Goodness_criterion
        {
            INFO_GAIN=0x00,
            GINI=0x01,

        }gc;

        /** @brief Used to pass information about which data to work with.**/                
        struct RowsSubIndexes
        {
            uint32_t size=0;
            uint32_t* indexes=NULL;
            ~RowsSubIndexes(){
                if(indexes)free(indexes);
            }
        };

    protected:

        struct UniqueValues
        {
            T* uniqueValues=NULL;
            uint32_t uniqueValuesSize=0;
            uint32_t* uniqueValuesOccurances=NULL;
            ~UniqueValues(){
                if(uniqueValues)free(uniqueValues);
                if(uniqueValuesOccurances)free(uniqueValuesOccurances);
            }
        }uniqueClassValuesAndOccurances;

            /** @brief Compute Shannon's entropy.
         * @param Y Input column.
         * @param rowsToProcess Which rows to process.
        **/                
        float computeEntropy(T** Y,RowsSubIndexes* rowsToProcess);

        /** @brief Compute Gini impurity entropy.
         * @param Y Input column.
         * @param rowsToProcess Which rows to process.
        **/                
        float computeGini(T** Y,RowsSubIndexes* rowsToProcess);

        /** @brief Returns the most frequent class.
         * @param uv A pointer to array of data to sort.
          **/                
        T getMajorClass(UniqueValues* uv);

        /** @brief Checks how many unique values does the columm have. Used to calculate Shannons entropy.
         * @param ar All the input data.
         * @param rsi Which rows to process.
         * @param column Which column to process.
         * @param uv A pointer to the output variable to fill in the data.
        **/                
        void countUniqueValuesAndOccurances(T** ar, RowsSubIndexes* rsi, uint32_t column, UniqueValues* uv);

        /** @brief Checks how many unique values does the columm have. Used to calculate Shannons entropy.
         * @param ar All the input data.
         * @param rsi Which rows to process.
         * @param column Which column to process.
         * @param uv A pointer to the output variable to fill in the data.
        **/                
        void countOccurances(T** ar, RowsSubIndexes* rsi, uint32_t column, UniqueValues* uv);

        /** @brief Checks If only one class is left in the column, should be called after counting.
         * @param uv A pointer to the output variable to fill in the data.
        **/                
        bool onlyClassLeft(UniqueValues* uv);

        /** @brief Quicksort is used for sorting the data to split the data by threholds. The data are sorted in place.
         * @param ar A pointer to array of data to sort.
         * @param n Array size
          **/                
        void qsort(T * ar,uint32_t n);

        /** @brief Quicksort is used for sorting the data to split the data by threholds. The data are sorted in place.
         * @param ar A pointer to array of data to sort.
         * @param n Array size
         * @param idx Pointer to an array of indices, that describes which index ended up where after the sort
          **/                
        void qsort(T * ar, uint32_t* idx,uint32_t n);

        /** @brief Generates the split that maximizes information gain. 
         * @param X Input samples.
         * @param Y Input classes.
         * @param rowsToProcess Which rows to process.
         * @param cols Total number of input samples columns (basically the number of input features).
         * @param rsiAboveTh Output variable, rows above threhold
         * @param rsiBelowTh Output variable, rows below threhold
         * @param threshold Output variable, a pointer to the best threhold.(returned after method ends)
         * @param column Output variable, a column of the best split(returned after method ends)
        **/                
        int16_t getBestSplitInfoGain(T** X,T** Y, RowsSubIndexes* rowsToProcess, uint32_t cols, RowsSubIndexes* rsiAboveTh, RowsSubIndexes* rsiBelowTh, T* threshold, uint32_t* column);

        /** @brief Generates the split that minimizes Gini impurity. 
         * @param X Input samples.
         * @param Y Input classes.
         * @param rowsToProcess Which rows to process.
         * @param cols Total number of input samples columns (basically the number of input features).
         * @param rsiAboveTh Output variable, rows above threhold
         * @param rsiBelowTh Output variable, rows below threhold
         * @param threshold Output variable, a pointer to the best threhold.(returned after method ends)
         * @param column Output variable, a column of the best split(returned after method ends)
        **/                
        int16_t getBestSplitGini(T** X,T** Y, RowsSubIndexes* rowsToProcess, uint32_t cols, RowsSubIndexes* rsiAboveTh, RowsSubIndexes* rsiBelowTh, T* threshold, uint32_t* column);

    public:
    uint16_t maxDepth;
    uint16_t minSamplesSplit;
    uint16_t cols;

    /** @brief Nested node class**/
    class Node{
        public:
   
        /** @brief Child nodes.**/                
        Node* children[2] = {NULL,NULL};
        bool decisionReady = false;
        T decision;
        struct NodeDecisionPair{
            T decision;
            Node* nodePtr=NULL;
        };

        bool thReady = false;
        T nodeTh;
        uint32_t nodeThColumn;

        /** @brief Recursively builds the tree, generating the best split by maximizing information gain.
         * @param minSamplesSplit If the number of samples is less than this number the splitting process stops.
         * @param maxDepth Maximum possible depth of a tree.
        **/                
        Node(void);

        /** @brief Recursively make a decision about the output class.
         * @param X Input sample.
        **/                
        T decide(T *X);

        /** @brief Recursively make a decision about the output class, but use a pre-leaf node decision instead of leaf. Used for cost complexity pruning.
         * @param X Input sample.
        **/                
        void decideBasedOnPreLeafNode(T *X,NodeDecisionPair* ndp);
        
        /** @brief Recursively destroys the nodes and deallocates the memory**/                
        void cleanup(void);
        };
        
        Node* root;

        /** @brief The class constructor.
         * @param minSamplesSplit If the number of samples is less than this number the splitting process stops.
         * @param maxDepth Maximum possible depth of a tree.
        **/                
        TinyDecisionTreeClassifier(uint16_t maxDepth, uint16_t minSamplesSplit = 2,Goodness_criterion gc=INFO_GAIN);

        /** @brief The class destructor.Cleans all the memory
        **/                
        ~TinyDecisionTreeClassifier();

        /** @brief Recursively plots the tree.
         * @param node Pointer to the root node.
         * @param depth Current depth.
        **/                
        void plot(Node* node,uint32_t depth);

        /** @brief Recursively plots the tree.**/                
        void plot();

        /** @brief Recursively builds the tree, generating the best split by maximizing information gain.
         * @param X Input samples.
         * @param Y Input classes.
         * @param rsi Indicies of rows to work with.
         * @param cols Total number of input samples columns (basically the number of input features).
         * @param current_depth Since the method is called recursively this argument tracks the current depth.
        **/                
        int16_t recurcisiveFit(Node* node, T** X,T** Y, RowsSubIndexes* rsi, uint32_t cols, uint32_t current_depth);

        /** @brief Fits the tree to input data.
         * @param X Input samples.
         * @param Y Input classes.
         * @param rows Number or samples.
         * @param Xcols Number of features.
         * **/                
        void fit(T** X,T** Y, uint32_t rows,uint32_t Xcols, RowsSubIndexes* rootRsi=NULL);

        /** @brief Classifies the input. The tree should be trained using fit method first. 
         * @param X Input samples.
         * **/                
        T predict(T* X); 

        /** @brief Checks the accuracy of trained tree. 
         * @param X Input samples.
         * @param Y Input features.
         * @param rows Number of samples
         * **/                
        float score(T** X,T** Y,uint32_t rows);

        /** @brief Classifies the data, but uses pre-leaf value instead of leaf value. Used for cost-complexity pruning.
         * @param X A pointer to array of data to classify.
         * @param ndp Decision and Node address pair.
          **/                
        T predictBasedOnPreLeafNode(T *X,typename Node::NodeDecisionPair* ndp);

        /** @brief Prunes the decision tree. Cost-complexity pruning is used 
         * @param X Input samples.
         * @param Y Input features.
         * @param rows Number of samples
         * @return True if at least one node was removed as a result of pruning, false otherwise
         * **/                
        bool prune(T** X,T** Y,uint32_t rows);

        /** @brief Saves the decision tree to file. 
         * @param outdata output array.
         * @param outdataMaxSize the maximum size that can be occupied.
         * **/                
        int save(char** outdata, uint32_t* outdataMaxSize);

        /** @brief Loads the decision tree from file. 
         * @param inputdata char array with the saved tree.
         * **/                
        int load(char* inputdata);

    private:
        int32_t getRootLine(Node* node,int32_t linenow);

        int32_t nodesToChar(char* outdata, uint32_t* outdataOffset, Node* node,int32_t linenow);

        void countNodesAndRoots(uint32_t* roots,uint32_t* nodes,char* data);

        void fillInNodesData(uint32_t* roots,uint32_t* nodes,char* inputdata,float** nodesData,float** rootsData);

        void buildDescendents(Node* ptr,int32_t index,float** nodesData);

};

#include "TinyDecisionTreeClassifier.cpp"

#endif